// gcc -Wall -O3 -march=native -I. -std=gnu99 -ggdb3 -I/usr/include/netpbm -o aim-opencl aim-opencl-run.c -lnetpbm -lOpenCL

#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <pbm.h>
#include <CL/cl.h>

#include "aim.h"


#if CHAR_BIT != 8
#error char size larger than 8 is not supported
#endif


/* XXX: This is a bit different from default aim-kit. */

static void
image_probe(char *filename, int *cols, int *rows, int *fmt)
{
	FILE *f = fopen(filename, "rb");
	if (!f) { perror("load"); exit(EXIT_FAILURE); }
	pbm_readpbminit(f, cols, rows, fmt);
	fclose(f);
}

static size_t
bitmap_size(int cols, int rows)
{
	return rows * (cols + 7) / 8;
}

static void
image_load(struct image *img, int fmt, char *filename)
{
	FILE *f = fopen(filename, "rb");
	if (!f) { perror("load"); exit(EXIT_FAILURE); }
	int i,j,k;
	pbm_readpbminit(f, &i, &j, &k);

	for (int i = 0; i < img->rows; i++)
		pbm_readpbmrow_packed(f, img->bitmap + i * image_rowbytes(img),
				img->cols, fmt);

	fclose(f);
}

static void
image_save(char *filename, struct image *img)
{
	FILE *f = fopen(filename, "wb");
	if (!f) { perror("save"); exit(EXIT_FAILURE); }

	pbm_writepbminit(f, img->cols, img->rows, 0);
	for (int i = 0; i < img->rows; i++)
		pbm_writepbmrow_packed(f, img->bitmap + i * image_rowbytes(img),
				img->cols, 0);

	fclose(f);
}


typedef int_fast64_t timestamp_t;

static timestamp_t
get_timer(void)
{
	struct timeval t;
	gettimeofday(&t, NULL);
	return 1000000*t.tv_sec + t.tv_usec;
}

static char *
file_contents(const char *filename)
{
	FILE *f = fopen(filename, "r");
	if (!f) {
		perror(filename);
		exit(EXIT_FAILURE);
	}

	fseek(f, 0, SEEK_END);
	int length = ftell(f);
	fseek(f, 0, SEEK_SET);

	char *buffer = malloc(length + 1);
	length = fread(buffer, 1, length, f);
	fclose(f);
	buffer[length] = '\0';

	return buffer;
}

static void
buildNotify(cl_program program, void *data)
{
	cl_device_id *device = data;
	cl_build_status status;
	clGetProgramBuildInfo(program, *device, CL_PROGRAM_BUILD_STATUS, sizeof(status), &status, NULL);
	char log[65536];
	clGetProgramBuildInfo(program, *device, CL_PROGRAM_BUILD_LOG, sizeof(log), &log, NULL);
	fprintf(stderr, "build status: %d, log:\n%s\n\n", status, log);
}


int
main(int argc, char *argv[])
{
	/* Parse parameters. */

	if (argc != 6) {
		fprintf(stderr, "%s ITERATIONS JOBS THREADSPERJOB SRCIMAGE.pbm DSTIMAGE.pbm\n", argv[0]);
		return EXIT_FAILURE;
	}

	int iters = atoi(argv[1]);
	size_t jobs = atoi(argv[2]);
	size_t tpj = atoi(argv[3]);
	char *srcimage = argv[4];
	char *dstimage = argv[5];


	/* Set up images. */

	int cols, rows, fmt;
	image_probe(srcimage, &cols, &rows, &fmt);
	size_t bitmapsize = bitmap_size(cols, rows);
	struct image img_in[jobs];
	unsigned char (*bitmap_in)[bitmapsize] = malloc(jobs * bitmapsize);
	for (int i = 0; i < jobs; i++) {
		img_in[i].cols = cols;
		img_in[i].rows = rows;
		img_in[i].bitmap = bitmap_in[i];
		image_load(&img_in[i], fmt, srcimage);
	}
	struct image img_out[jobs];
	unsigned char (*bitmap_out)[bitmapsize] = malloc(jobs * bitmapsize);
	memset(bitmap_out, 0xca, jobs * bitmapsize); // characteristic pattern
	for (int i = 0; i < jobs; i++) {
		img_out[i].cols = cols;
		img_out[i].rows = rows;
		img_out[i].bitmap = bitmap_out[i];
	}


	/* Set up OpenCL. */

	// Get OpenCL platform count
	cl_uint NumPlatforms;
	clGetPlatformIDs(0, NULL, &NumPlatforms);
	
	// Get all OpenCL platform IDs
	cl_platform_id *PlatformIDs = malloc(sizeof(cl_platform_id) * NumPlatforms);
	clGetPlatformIDs(NumPlatforms, PlatformIDs, NULL);

	// Select NVIDIA platform (this example assumes it IS present)
	/* TODO: Better way to do this. */
	char cBuffer[1024];
	cl_uint NvPlatform = 0;
	for(cl_uint i = 0; i < NumPlatforms; ++i)
	{
		clGetPlatformInfo (PlatformIDs[i], CL_PLATFORM_NAME, 1024, cBuffer, NULL);
		if(strstr(cBuffer, "NVIDIA") != NULL)
		{
			NvPlatform = i;
			break;
		}
	}

	// Get a GPU device on Platform (this example assumes one IS present)
	cl_device_id cdDevice;
	clGetDeviceIDs(PlatformIDs[NvPlatform], CL_DEVICE_TYPE_GPU, 1,
			&cdDevice, NULL);
	// Create a context
	cl_context hContext;
	hContext = clCreateContext(0, 1, &cdDevice, NULL, NULL, NULL);
	// Create a command queue for the device in the context
	cl_command_queue hCmdQueue;
	hCmdQueue = clCreateCommandQueue(hContext, cdDevice, 0, NULL);
	// Create & compile program
	const char *sProgramSource = file_contents("exercise.cl"); // TODO: Commandline parameter.
	cl_program hProgram;
	hProgram = clCreateProgramWithSource(hContext, 1, &sProgramSource, 0, 0);
	clBuildProgram(hProgram, 0, NULL, "-I.", buildNotify, &cdDevice);
	// Create kernel instance
	cl_kernel hKernel;
	hKernel = clCreateKernel(hProgram, "exercise", 0);


	timestamp_t t0 = get_timer();

	/* Pre-warm the CPU! Neccessary for benchmarking w/ dynamic
	 * cpufreq policy. */
	while (get_timer() - t0 < 500000);
	t0 = get_timer();

	for (int j = 0; j < iters; j++) {
		/* TODO: Detailed timing information within this loop. */
		/* TODO: Different work partitioning - single job, various
		 * workgroups divide huge picture to blocks they chew on. */
		// Allocate CL memory for the image
		cl_mem cimg_in = clCreateBuffer(hContext, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, sizeof(struct image) * jobs, img_in, NULL);
		cl_mem cbitmap_in = clCreateBuffer(hContext, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, bitmapsize * jobs, bitmap_in, NULL);
		cl_mem cimg_out = clCreateBuffer(hContext, CL_MEM_WRITE_ONLY, sizeof(struct image) * jobs, NULL, NULL);
		cl_mem cbitmap_out = clCreateBuffer(hContext, CL_MEM_WRITE_ONLY, bitmapsize * jobs, NULL, NULL);
		// Setup parameter values
		clSetKernelArg(hKernel, 0, sizeof(cl_mem), &cimg_in);
		clSetKernelArg(hKernel, 1, sizeof(cl_mem), &cbitmap_in);
		clSetKernelArg(hKernel, 2, sizeof(cl_mem), &cimg_out);
		clSetKernelArg(hKernel, 3, sizeof(cl_mem), &cbitmap_out);
		clSetKernelArg(hKernel, 4, sizeof(cl_uint), &bitmapsize);
		// Launch kernel
		size_t tot = jobs * tpj;
		int ret = clEnqueueNDRangeKernel(hCmdQueue, hKernel, 1, NULL, &tot, &tpj, 0, NULL, NULL);
		if (ret != CL_SUCCESS) {
			fprintf(stderr, "kernel failed: %d\n", ret);
		}
		// Copy results from device back to host; block until complete
		clEnqueueReadBuffer(hCmdQueue, cimg_out, CL_TRUE, 0,
		                    sizeof(struct image) * jobs, img_out, 0, NULL, NULL);
		clEnqueueReadBuffer(hCmdQueue, cbitmap_out, CL_TRUE, 0,
		                    bitmapsize * jobs, bitmap_out, 0, NULL, NULL);
		// Release CL memory
		clReleaseMemObject(cimg_in);
		clReleaseMemObject(cbitmap_in);
		clReleaseMemObject(cimg_out);
		clReleaseMemObject(cbitmap_out);
	}

	t0 = get_timer() - t0;
	// time spent:
	printf("%.3f\n", (double) t0/1e6);


	clReleaseKernel(hKernel);
	clReleaseProgram(hProgram);
	clReleaseCommandQueue(hCmdQueue);
	clReleaseContext(hContext);
	free(PlatformIDs);
	free((void *) sProgramSource);

	img_out[jobs - 1].bitmap = bitmap_out[jobs - 1];
	image_save(dstimage, &img_out[jobs - 1]);
	free(bitmap_out);
	free(bitmap_in);
	return EXIT_SUCCESS;
}
